-- Fonction permettant d'encapsuler les contrôles de cohérence et de conformité sur les données 
-- à lancer après l'import d'un jeu de données pour s'assurer qu'il n'y ait pas de problème.
-- Le script exécute une batterie de tests puis renvoie une table listant les anomalies


-- 1. Création préalable de la table divers.controle_coherence_conformite si besoin. 
----------------------------------------------------------------------------------
-- Cette table permet de stocker le résultat de la fonction
--  DROP TABLE divers.controle_coherence_conformite;
CREATE TABLE IF NOT EXISTS divers.controle_coherence_conformite
(
  jdd_id TEXT NOT NULL,
  jdd_code text,
  cle_obs integer,
  identifiant_origine text, 
  wkt text,
  libelle_test text,
  description_anomalie text,
  date_analyse timestamp with time zone,
  nom_cite TEXT,
  nom_valide TEXT,
  reu TEXT,
  habitat TEXT,
  CONSTRAINT fdw_controle_coherence_conformite_pk PRIMARY KEY (cle_obs, libelle_test)
)
WITH (
  OIDS=FALSE
);
ALTER TABLE divers.controle_coherence_conformite
  OWNER TO postgres;
  
GRANT ALL ON TABLE divers.controle_coherence_conformite TO lizmap;
GRANT ALL ON TABLE divers.controle_coherence_conformite TO postgres;
  
COMMENT ON TABLE divers.controle_coherence_conformite
  IS 'Table listant les anomalies issues de l''analyse de contrôle et de conformité des tables testées';
COMMENT ON COLUMN divers.controle_coherence_conformite.jdd_id IS 'jdd_id du jeu de données de l''observation';
COMMENT ON COLUMN divers.controle_coherence_conformite.jdd_code IS 'jdd_code du jeu de données de l''observation';
COMMENT ON COLUMN divers.controle_coherence_conformite.cle_obs IS 'cle_obs de l''observation';
COMMENT ON COLUMN divers.controle_coherence_conformite.identifiant_origine IS 'identifiant_origine de l''observation';
COMMENT ON COLUMN divers.controle_coherence_conformite.libelle_test IS 'Libellé du test';
COMMENT ON COLUMN divers.controle_coherence_conformite.description_anomalie IS 'Description détaillée de l''anomalie détectée par le test';
COMMENT ON COLUMN divers.controle_coherence_conformite.date_analyse IS 'Date à laquelle l''analyse a été réalisée';
COMMENT ON COLUMN divers.controle_coherence_conformite.nom_cite IS 'Nom de taxon cité';
COMMENT ON COLUMN divers.controle_coherence_conformite.nom_valide IS 'Nom valide du taxon issu de Taxref';
COMMENT ON COLUMN divers.controle_coherence_conformite.reu IS 'Statut biogéographique issu de Taxref';
COMMENT ON COLUMN divers.controle_coherence_conformite.habitat IS 'Habitat du taxon d''après Taxref';





-- 2. Création de la fonction divers.fonction_controle_coherence_conformite(text)
-----------------------------------------------------------------------
-- DROP FUNCTION divers.fonction_controle_coherence_conformite;

CREATE OR REPLACE FUNCTION divers.fonction_controle_coherence_conformite(jdd_id text[])
  RETURNS void
AS $BODY$ 

DECLARE liste_champs RECORD ;

BEGIN

-- a/ Vidage de la table
DELETE FROM divers.controle_coherence_conformite ;

-- b/ remplissage des valeurs uniques depuis la table testée

WITH o AS (
		SELECT 	o.jdd_id,o.jdd_code, o.cle_obs, o.identifiant_origine, o.nom_cite, o.cd_nom, o.ds_publique, o.diffusion_niveau_precision, tc.cd_nom AS cd_nom_taxref, t.cd_ref, t.nom_valide, t.reu, t.habitat, CONCAT(t.habitat, ' - ', h.valeur) AS habitat_decode, 
		o.geom, lc.code_commune, ltm.geom AS sur_terre, min(r.larg_pb_th) AS en_riviere, zee.geom AS dans_zee, 
				COALESCE(altitude_min, altitude_moy,altitude_max) AS altitude, profondeur_min
		FROM occtax.observation o
		LEFT JOIN taxon.taxref_consolide_non_filtre tc ON o.cd_nom = tc.cd_nom
		LEFT JOIN taxon.taxref_valide t ON o.cd_ref=t.cd_ref
		LEFT JOIN (SELECT * FROM taxon.t_nomenclature WHERE champ= 'habitat') h ON h.code = t.habitat
		LEFT JOIN sig.limite_terre_mer ltm ON st_intersects(ltm.geom, o.geom)
		LEFT JOIN sig.riviere r ON st_distance(r.geom, o.geom)::INTEGER <= r.larg_pb_th::INTEGER
		LEFT JOIN sig.zone_economique_exclusive zee ON st_intersects(zee.geom, o.geom)
		LEFT JOIN occtax.localisation_commune lc ON lc.cle_obs = o.cle_obs
		WHERE o.jdd_id = ANY ($1)
		GROUP BY o.jdd_id,o.jdd_code, o.cle_obs, o.identifiant_origine, o.nom_cite, o.cd_nom, o.ds_publique, o.diffusion_niveau_precision, tc.cd_nom, t.cd_ref, t.nom_valide,t.reu, t.habitat, o.geom, lc.code_commune, ltm.geom, zee.geom, 
				COALESCE(altitude_min, altitude_moy,altitude_max), profondeur_min, t.habitat, h.valeur
			)
			
	, test_5 AS (
	-- TEST 5 "Diffusion au grand public" : on vérifie que les modalités de diffusion au grand public sont bien précisées et sont compatibles avec le statut public de la donnée
	SELECT o.cle_obs
	FROM o
	WHERE 	(ds_publique = ANY (ARRAY['NSP'::text, 'Pr'::text]) AND diffusion_niveau_precision IS  NULL)
			OR (ds_publique = ANY (ARRAY['Pu'::text, 'Re'::text, 'Ac'::text])  AND diffusion_niveau_precision <> '5')

	ORDER BY o.cle_obs
			
	), test_11 AS (
	-- TEST 11 "Altitude des observations marines" : on vérifie que les observations concernant un organisme marin n'ont pas une altitude supérieure à 0m
	SELECT o.cle_obs
	FROM o
	WHERE altitude>0 AND habitat='1'
	ORDER BY o.cle_obs
	
	), test_12 AS (
	-- TEST 12 "Profondeur des observations terrestres"	: on vérifie que les observations concernant un organisme terrestre n'ont pas une profondeur supérieure à 0m
	SELECT o.cle_obs
	FROM o
	WHERE profondeur_min>0 AND habitat='3'
	ORDER BY o.cle_obs

	), test_13 AS (		
	-- TEST 13 "localisation des espèces marines" : on vérifie que les observations concernant un organisme marin n'ont pas lieu en milieu terrestre
	SELECT o.cle_obs
	FROM o
	WHERE habitat='1' AND o.sur_terre IS NOT NULL
	ORDER BY o.cle_obs
		
	), test_14 AS (
	-- TEST 14 "localisation des espèces terrestres" : on vérifie que les observations concernant un organisme terrestre n'ont pas lieu en mer
	SELECT o.cle_obs
	FROM o
	WHERE habitat IN ('2', '3', '7', '8') AND o.sur_terre IS NULL AND o.code_commune IS NULL
	ORDER BY o.cle_obs
	
	-- test 15 : "localisation des espèces d'eau douce" : on vérifie que les observations de taxons d'eau douce n'ont pas lieu en dehors d'un cours d'eau
	), test_15 AS (
	SELECT o.cle_obs
	FROM o
	WHERE habitat IN ('2') AND o.en_riviere IS NULL
	ORDER BY o.cle_obs
	
	), test_16 AS (
	-- TEST 16 "identifiants du jeu de données non cohérents" : on vérifie que les jdd_id et jdd_code sont les mêmes entre les tables occtax et jdd
	SELECT o.cle_obs
	FROM o
	WHERE o.jdd_id NOT IN (SELECT DISTINCT jdd.jdd_id FROM occtax.jdd) OR o.jdd_code NOT IN (SELECT DISTINCT jdd.jdd_code FROM occtax.jdd)
	ORDER BY o.cle_obs	
		
	),test_27 AS (
	-- TEST 27 "champ obligatoire : identite" : On vérifie qu'on a bien au moins un observateur/observation
	SELECT o.cle_obs
	FROM o
	LEFT JOIN occtax.observation_personne op USING(cle_obs)
	WHERE op.id_personne IS NULL
	ORDER BY o.cle_obs

	), test_41 AS (
	-- TEST 41 "champ obligatoire : rattachement géographique" : on vérifie que chaque observation est associée à exactement une géométrie de référence
	-- Il peut s'agir d'une géométrie précise, ou bien d'un espace de référence (maille, commune...). D'autres géométries de rattachement peuvent être associées.
	SELECT o.cle_obs,
		st_summary(o.geom),
		count(DISTINCT o.geom) AS nb_geom, 
		count(DISTINCT lc.code_commune) AS nb_commune, 
		count(DISTINCT lm02.code_maille) AS nb_maille_02, 
		count(DISTINCT lm10.code_maille) AS nb_maille_10,
		count(DISTINCT lme.code_me) AS nb_masse_eau,	
		count(DISTINCT len.code_en) AS nb_espace_nat,	
		count(DISTINCT ld.code_departement) AS nb_departement
	FROM o
	LEFT JOIN occtax.localisation_commune lc ON lc.cle_obs=o.cle_obs
	LEFT JOIN occtax.localisation_maille_02 lm02 ON lm02.cle_obs=o.cle_obs
	LEFT JOIN occtax.localisation_maille_02 lm10 ON lm10.cle_obs=o.cle_obs
	LEFT JOIN occtax.localisation_masse_eau lme ON lme.cle_obs=o.cle_obs
	LEFT JOIN occtax.localisation_espace_naturel len ON len.cle_obs=o.cle_obs
	LEFT JOIN occtax.localisation_departement ld ON ld.cle_obs=o.cle_obs
	WHERE 
		(o.geom IS NULL AND CONCAT(COALESCE(lc.type_info_geo,''), 
								   COALESCE(lm02.type_info_geo,''), 
								   COALESCE(lm10.type_info_geo,''), 
								   COALESCE(lme.type_info_geo,''),
								   COALESCE(len.type_info_geo,''),
								   COALESCE(ld.type_info_geo,''))  NOT ILIKE '%1%') -- Aucune géométrie de référence

		OR	(o.geom IS NULL AND CONCAT(COALESCE(lc.type_info_geo,''), 
								   COALESCE(lm02.type_info_geo,''), 
								   COALESCE(lm10.type_info_geo,''), 
								   COALESCE(lme.type_info_geo,''),
								   COALESCE(len.type_info_geo,''),
								   COALESCE(ld.type_info_geo,''))  ILIKE '%1%1%') -- Plusieurs géométries de référence sans geom précise
											
		OR (o.geom IS NOT NULL AND CONCAT(COALESCE(lc.type_info_geo,''), 
								   COALESCE(lm02.type_info_geo,''), 
								   COALESCE(lm10.type_info_geo,''), 
								   COALESCE(lme.type_info_geo,''),
								   COALESCE(len.type_info_geo,''),
								   COALESCE(ld.type_info_geo,'')) ILIKE '%1%') -- Plusieurs géométries de référence dont géom précise
	GROUP BY o.cle_obs, o.geom, o.jdd_code, o.jdd_id, o.identifiant_origine
	ORDER BY cle_obs
	
	), test_46 AS (
	-- TEST 46 "localisation des espèces dans la ZEE" : on vérifie que les observations ne sont pas situées en dehors de la ZEE
	SELECT  o.cle_obs
	FROM o
	WHERE o.dans_zee IS NULL AND o.geom IS NOT NULL
	ORDER BY o.cle_obs	

	), test_47 AS (
	-- TEST 47 "champs obligatoires : cd_nom et cd_ref" : on vérifie que les observations sont bien associées à un cd_nom et un cd_ref issus de Taxref ou Taxref_local
	SELECT  o.cle_obs
	FROM o
	WHERE o.cd_nom_taxref IS NULL or o.cd_ref IS NULL
	ORDER BY o.cle_obs		
	
	), liste AS (	
-- On fusionne les différentes tables pour avoir une table globale des anomalies
	SELECT 	cle_obs,
			'test_5' AS libelle_test, 
			'Les modalités de diffusion au grand public ne sont pas précisées ou bien sont incompatibles avec le statut public de la donnée' AS description_anomalie
	FROM test_5
	UNION
	SELECT 	cle_obs,
			'test_11' AS libelle_test, 
			'L''observation concerne un organisme marin et est à une altitude supérieure à 0m' AS description_anomalie
	FROM test_11
	UNION
	SELECT 	cle_obs,
			'test_12' AS libelle_test, 
			'L''observation concerne un organime terrestre et est à une profondeur supérieure à 0m'  AS description_anomalie
	FROM test_12
	UNION
	SELECT cle_obs ,
			'test_13' AS libelle_test, 
			'L''observation concerne un organisme marin mais a lieu en milieu terrestre'  AS description_anomalie
	FROM test_13
	UNION
	SELECT cle_obs,
			'test_14' AS libelle_test, 
			'L''observation concerne un organisme terrestre mais a lieu en mer'  AS description_anomalie
	FROM test_14
	UNION
	SELECT 	cle_obs, 
			'test_15' AS libelle_test, 
			'L''observation concerne un organisme d''eau douce mais a lieu en dehors d''un cours d''eau'  AS description_anomalie
	FROM test_15
	UNION
	SELECT 	cle_obs, 
			'test_16' AS libelle_test, 
			'Les identifiants du jeu de données jdd_id et jdd_code ne sont pas cohérents entre la table observation et la table jdd'  AS description_anomalie
	FROM test_16
	UNION
	SELECT 	cle_obs,
			'test_27' AS libelle_test, 
			'L''observation est orpheline : elle n''est associée à aucun observateur' AS description_anomalie
	FROM test_27
	UNION
	SELECT cle_obs,
			'test_41' AS libelle_test, 
			'L''observation n''est pas associée à un unique objet géographique de référence' AS description_anomalie
	FROM test_41
	UNION
	SELECT 	cle_obs, 
			'test_46' AS libelle_test, 
			'L''observation est située en dehors de la ZEE' AS description_anomalie
	FROM test_46
	UNION
	SELECT 	cle_obs, 
			'test_47' AS libelle_test, 
			'L''observation n’est pas associée à un cd_nom ou un cd_ref issu de Taxref ou Taxref_local' AS description_anomalie
	FROM test_47
	)

INSERT INTO divers.controle_coherence_conformite
(jdd_id, jdd_code, cle_obs, identifiant_origine, wkt, libelle_test, description_anomalie, date_analyse, nom_cite, nom_valide, reu, habitat )
SELECT o.jdd_id, o.jdd_code, l.cle_obs, o.identifiant_origine, st_AsEWKT(o.geom), l.libelle_test, l.description_anomalie, now()::TIMESTAMP WITH TIME ZONE, o.nom_cite, o.nom_valide, o.reu, o.habitat_decode
FROM liste l
LEFT JOIN o USING(cle_obs)
ORDER BY l.libelle_test, o.cle_obs
;
										 
END ;
										 
$BODY$
  LANGUAGE plpgsql VOLATILE
  COST 100;
ALTER FUNCTION divers.fonction_controle_coherence_conformite(text[])
  OWNER TO postgres;
										 
										 
-- 3. On teste la fonction
--------------------------------
SELECT divers.fonction_controle_coherence_conformite(
	(SELECT array_agg(jdd_id) FROM occtax.jdd WHERE jdd_code='ecomedoi_faune_historique')
	);
SELECT * FROM divers.controle_coherence_conformite ;
								